import os
import pkg_resources

controller_temp = pkg_resources.resource_filename('renus', './commands/app/controller.temp')
model_temp = pkg_resources.resource_filename('renus', './commands/app/model.temp')
vue_temp = pkg_resources.resource_filename('renus', './commands/app/vue.temp')
route_temp = pkg_resources.resource_filename('renus', './commands/app/route.temp')

def to_camel_case(snake_str):
    components = snake_str.split('_')
    return ''.join(x.title() for x in components)


def add_temp(name,names, name_camel, name_db,folder:list):
    pth = 'app/'
    pname = ''
    vname = ''
    if len(folder) > 0:
        pth += '/'.join(folder) + '/'
        pname = '.'.join(folder) + '.'
        vname = '/'+'/'.join(folder)+'/'
    pth += name + '/'
    with open(controller_temp,'r') as file:
        os.makedirs(pth,exist_ok=True)
        all=file.read()
        all=all.replace('{name}',pname+name)
        all=all.replace('{name_camel}',name_camel)
        with open(pth+'controller.py','w') as controller:
            controller.write(all)

    with open(model_temp,'r') as file:
        all=file.read()
        all=all.replace('{name_camel}',name_camel)
        all=all.replace('{name_db}',name_db)
        with open(pth+'model.py','w') as controller:
            controller.write(all)

    with open(vue_temp,'r') as file:
        os.makedirs('./frontend/src/views/admin/'+vname.lstrip('/'), exist_ok=True)
        all=file.read()
        all=all.replace('{name}',name)
        all=all.replace('{names}',names)
        all = all.replace('{folder}', vname.lstrip('/'))
        with open('./frontend/src/views/admin/'+vname.lstrip('/')+name+'.vue','w') as controller:
            controller.write(all)

    with open(route_temp,'r') as file:
        all=file.read()
        all=all.replace('{name}',pname+name)
        all=all.replace('{names}',names)
        all = all.replace('{name_camel}', name_camel)
        all = all.replace('{folder}', vname.rstrip('/'))
        with open(pth+'route.py','w') as controller:
            controller.write(all)

def to_plural(word):
    slice = list(word)
    if slice[-1] == 'y':
        slice.remove('y')
        slice.append('ies')

    elif slice[- 1] == 'f':
        slice.remove('f')
        slice.append('ves')

    elif slice[- 2:] == ['f','e']:
        slice.remove('fe')
        slice.append('ves')

    elif slice[- 1] == 's' or slice[-1] == 'x' or slice[-1] == 'z' or slice[- 2:] == ['c','h'] or slice[- 2:] == ['s','h']:
        slice.append('es')

    elif slice[- 2:] == ['u','s']:
        slice.remove('us')
        slice.append('i')

    else:
        slice.append('s')

    return ''.join(slice)


def add_routes(name, names, name_camel,folder):
    pname = ''
    vname = ''
    if len(folder) > 0:
        pname = '.'.join(folder) + '.'
        vname = '/'.join(folder) + '/'

    with open('./routes/index.py', 'r') as file:
        all=file.read()
        all=f'import app.{pname+name}.route\n'+all

        with open('./routes/index.py', 'w') as b:
                b.write(all)

    with open('./frontend/src/router/admin.js', 'r') as file:
        all=file.read()
        all=all.replace('/* {{place new import}} */','/* {{place new import}} */\nconst '+names+' = () => import("../views/admin/'+vname+name+'");')
        all=all.replace('/* {{place new Route}} */','/* {{place new Route}} */\n{ path: "'+vname+names+'", name: "'+names+'", component: '+names+' },')

        with open('./frontend/src/router/admin.js', 'w') as b:
                b.write(all)


def run():
    name = input('name (single lowercase):').lower()
    folder=name.split('.')
    name=folder[-1]
    folder=folder[:-1]
    name_camel = to_camel_case(name)
    names = to_plural(to_camel_case(name))
    names = names[0].lower() + names[1:]
    name_db = name.replace('_', '-')

    add_temp(name, names, name_camel, name_db,folder)
    print('templates created.')
    add_routes(name, names, name_camel,folder)
    print('routes created.')
    print(f'{name} is done.')



import asyncio
import functools
import typing
import contextvars
from typing import Any, AsyncGenerator, Iterator

T = typing.TypeVar("T")


async def run_until_first_complete(*args: typing.Tuple[typing.Callable, dict]) -> None:
    tasks = [handler(**kwargs) for handler, kwargs in args]
    (done, pending) = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
    [task.cancel() for task in pending]
    [task.result() for task in done]


async def run_in_threadpool(
    func: typing.Callable[..., T], *args: typing.Any, **kwargs: typing.Any
) -> T:
    loop = asyncio.get_event_loop()
    # Ensure we run in the same context
    child = functools.partial(func, *args, **kwargs)
    context = contextvars.copy_context()
    func = context.run
    args = (child,)
    return await loop.run_in_executor(None, func, *args)


class _StopIteration(Exception):
    pass


def _next(iterator: Iterator) -> Any:
    # We can't raise `StopIteration` from within the threadpool iterator
    # and catch it outside that context, so we coerce them into a different
    # exception type.
    try:
        return next(iterator)
    except StopIteration:
        raise _StopIteration


async def iterate_in_threadpool(iterator: Iterator) -> AsyncGenerator:
    while True:
        try:
            yield await run_in_threadpool(_next, iterator)
        except _StopIteration:
            break



from hashlib import md5
import io
import json
import os
import stat
import http.cookies
import typing
import inspect
import aiofiles
import gzip
from aiofiles.os import stat as aio_stat
from email.utils import formatdate
from mimetypes import guess_type
from urllib.parse import quote, quote_plus

from renus.core.cache import Cache
from renus.core.serialize import jsonEncoder
from renus.core.status import Status
from renus.core.concurrency import iterate_in_threadpool, run_until_first_complete
from renus.core.datastructures import Background
from renus.util.helper import get_random_string
from renus.core.crypt import shift

class Response:
    media_type = "text/html"
    charset = "utf-8"

    def __init__(
            self,
            content: typing.Any = None,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            media_type: str = None,
            background: Background = None,
    ) -> None:
        self.status_code = status_code
        if media_type is not None:
            self.media_type = media_type
        self.body = self.render(content)
        self.init_headers(headers)
        self.background = background

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        return content.encode(self.charset)

    def init_headers(self, headers: typing.Mapping[str, str] = None) -> None:
        populate_content_type = True
        raw_headers = []
        if headers is not None:
            for k, v in headers.items():
                if k.lower()!='content-length':
                    raw_headers.append((k.lower().encode("utf-8"), v.encode("utf-8")))

                if k.lower()=='content-type':
                    populate_content_type=False

        content_type = self.media_type
        if content_type is not None and populate_content_type:
            if content_type.startswith("text/"):
                content_type += "; charset=" + self.charset
            raw_headers.append((b"content-type", content_type.encode("utf-8")))
        
        self.raw_headers = raw_headers

    def set_cookie(
            self,
            key: str,
            value: str = "",
            max_age: int = None,
            expires: int = None,
            path: str = "/",
            domain: str = None,
            secure: bool = False,
            httponly: bool = False,
            samesite: str = "lax",
    ) -> None:
        cookie = http.cookies.SimpleCookie()  # type: http.cookies.BaseCookie
        cookie[key] = value
        if max_age is not None:
            cookie[key]["max-age"] = max_age
        if expires is not None:
            cookie[key]["expires"] = expires
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be either 'strict', 'lax' or 'none'"
            cookie[key]["samesite"] = samesite
        cookie_val = cookie.output(header="").strip()
        self.raw_headers.append((b"set-cookie", cookie_val.encode("utf-8")))

    def delete_cookie(self, key: str, path: str = "/", domain: str = None) -> None:
        self.set_cookie(key, expires=0, max_age=0, path=path, domain=domain)

    async def __call__(self, request,scope, receive, send) -> None:
        headers=request.headers
        if "gzip" in headers.get("accept-encoding", "") and len(self.body) > 500:
            self.gzip_buffer = io.BytesIO()
            self.gzip_file = gzip.GzipFile(mode="wb", fileobj=self.gzip_buffer)
            self.gzip_file.write(self.body)
            self.gzip_file.close()
            self.body = self.gzip_buffer.getvalue()
            self.raw_headers.append((b"content-encoding", "gzip".encode("utf-8")))
        self.raw_headers.append((b"content-length", str(len(self.body)).encode("utf-8")))
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        await send({"type": "http.response.body", "body": self.body})

        if self.background is not None:
            await self.background()


class HtmlResponse(Response):
    media_type = "text/html"


class TextResponse(Response):
    media_type = "text/plain"

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        return str(content).encode(self.charset)

class JsonResponse(Response):
    media_type = "application/json"
    def render(self, content: typing.Any) -> bytes:
        if isinstance(content, bytes):
            return content

        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=True,
            indent=None,
            separators=(",", ":"),
            cls=jsonEncoder
        ).encode("utf-8")

class JsonResponseRedirect(Response):

    def __init__(self, content: typing.Any = None, headers: dict = None,
                 background: Background = None) -> None:
        super().__init__(content, 307, headers, "application/json", background)


    def render(self,url:str) -> bytes:
        content = {"location":quote_plus(url, safe=":/%#?&=@[]!$&'()*+,;")}

        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
            cls=jsonEncoder
        ).encode("utf-8")

class EncryptResponse(Response):
    media_type = "text/plain"

    def __init__(self, content: typing.Any ,password:str, status_code: Status = Status.HTTP_200_OK, headers: dict = None,
                 media_type: str = None, background: Background = None) -> None:
        self.password = password
        if headers is None:
            headers = {}

        headers["encrypted"] = '1'
        headers["encrypted_type"] = str(type(content))
        super().__init__(content, status_code, headers, media_type, background)

    def render(self, content: typing.Any) -> bytes:
        if isinstance(content, bytes):
            return content

        return shift(str(content),self.password).encode("utf-8")


class RedirectResponse(Response):
    def __init__(
            self,
            url: str,
            status_code: Status = Status.HTTP_307_TEMPORARY_REDIRECT,
            headers: dict = None,
            background: Background = None,
    ) -> None:
        if headers is None:
            headers = {}

        headers["location"] = quote_plus(url, safe=":/%#?&=@[]!$&'()*+,;")
        super().__init__(
            content=b"", status_code=status_code, headers=headers,background=background
        )


class StreamingResponse(Response):
    def __init__(
            self,
            content: typing.Any,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            media_type: str = None,
            background: Background = None,
    ) -> None:
        super().__init__()
        if inspect.isasyncgen(content):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type
        self.background = background
        self.init_headers(headers)

    async def listen_for_disconnect(self, receive) -> None:
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                break

    async def stream_response(self,request,scope, send) -> None:
        headers=request.headers
        is_zip=False
        if "gzip" in headers.get("accept-encoding", ""):
            is_zip = True
            self.gzip_buffer = io.BytesIO()
            self.gzip_file = gzip.GzipFile(mode="wb", fileobj=self.gzip_buffer)
            self.raw_headers.append((b"content-encoding", "gzip".encode("utf-8")))

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for chunk in self.body_iterator:
            if is_zip:
                self.gzip_file.write(chunk)
                chunk = self.gzip_buffer.getvalue()
                self.gzip_buffer.seek(0)
                self.gzip_buffer.truncate()

            if not isinstance(chunk, bytes):
                chunk = chunk.encode(self.charset)

            await send({"type": "http.response.body", "body": chunk, "more_body": True})
        chunk=b""
        if is_zip:
            self.gzip_file.write(chunk)
            self.gzip_file.close()
            chunk = self.gzip_buffer.getvalue()
            self.gzip_buffer.seek(0)
            self.gzip_buffer.truncate()
        await send({"type": "http.response.body", "body":chunk, "more_body": False})

    async def __call__(self,request, scope, receive, send) -> None:
        await run_until_first_complete(
            (self.stream_response, {"request":request,"scope":scope,"send": send}),
            (self.listen_for_disconnect, {"receive": receive}),
        )
        if self.background is not None:
            await self.background()


class FileResponse(Response):
    chunk_size = 4096

    def __init__(
            self,
            path: str,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            media_type: str = None,
            background: Background = None,
            filename: str = None,
            stat_result: os.stat_result = None,
            method: str = None,
            zipped:bool=True,
            on_end=None
    ) -> None:
        super().__init__()
        self.path = path
        self.status_code = status_code
        self.background = background
        self.filename = filename
        self.on_end=on_end
        self.zipped=zipped
        self.send_header_only = method is not None and method.upper() == "HEAD"
        if media_type is None:
            media_type = guess_type(filename or path)[0] or "text/plain"
        self.media_type = media_type
        self.init_headers(headers)
        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = "attachment; filename*=utf-8''{}".format(
                    content_disposition_filename
                )
            else:
                content_disposition = 'attachment; filename="{}"'.format(self.filename)
            self.raw_headers.append((b"content-disposition", content_disposition.lower().encode("utf-8")))

        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result,False)

    def set_stat_headers(self, stat_result: os.stat_result,is_zip) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = md5(etag_base.encode()).hexdigest()
        if not is_zip:
            self.raw_headers.append((b"content-length", content_length.lower().encode("utf-8")))
        self.raw_headers.append((b"last-modified", last_modified.lower().encode("utf-8")))
        self.raw_headers.append((b"etag", etag.lower().encode("utf-8")))

    async def __call__(self,request, scope, receive, send) -> None:
        headers=request.headers
        is_zip = False
        if self.zipped and "gzip" in headers.get("accept-encoding", "") and not self.send_header_only:
            is_zip = True
            self.gzip_buffer = io.BytesIO()
            self.gzip_file = gzip.GzipFile(mode="wb", fileobj=self.gzip_buffer)
            self.raw_headers.append((b"content-encoding", "gzip".encode("utf-8")))

        if self.stat_result is None:
            try:
                stat_result = await aio_stat(self.path)
                self.set_stat_headers(stat_result,is_zip)
            except FileNotFoundError:
                raise RuntimeError(f"File at path {self.path} does not exist.")
            else:
                mode = stat_result.st_mode
                if not stat.S_ISREG(mode):
                    raise RuntimeError(f"File at path {self.path} is not a file.")
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        if self.send_header_only:
            await send({"type": "http.response.body", "body": b"", "more_body": False})
        else:
            async with aiofiles.open(self.path, mode="rb") as file:
                more_body = True
                while more_body:
                    chunk = await file.read(self.chunk_size)
                    more_body = len(chunk) == self.chunk_size
                    if is_zip:
                        self.gzip_file.write(chunk)
                        if not more_body:
                            self.gzip_file.close()
                        chunk = self.gzip_buffer.getvalue()
                        self.gzip_buffer.seek(0)
                        self.gzip_buffer.truncate()

                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )
        if self.on_end is not None:
            self.on_end()
        if self.background is not None:
            await self.background()


class PrivateResponse(Response):
    media_type = "text/plain"
    def __init__(
            self,
            url: str,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            background: Background = None
    ) -> None:
        super().__init__(
            content=url, status_code=status_code, headers=headers,background=background
        )

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        rnd = get_random_string(50)
        Cache().put(rnd, 'access to private')
        content=f'/storage/private/{rnd}/{content}'
        return content.encode(self.charset)



import copy

from renus.core.exception import abort_if
from renus.util.helper import dictAttribute
import renus.core.validation.rules as vr


def keys_exists(element, keys):
    '''
    Check if *keys (nested) exists in `element` (dict).
    And create and set key None if not Exists
    '''
    if not isinstance(element, dict):
        raise AttributeError('keys_exists() expects dict as first argument.')
    if len(keys) == 0:
        raise AttributeError('keys_exists() expects at least two arguments, one given.')

    _element = element
    for i,key in enumerate(keys):
        try:
            _element = _element[key]
        except KeyError:
            _element[key]=None if i==len(keys)-1 else {}
            if i<len(keys):
                _element =_element[key]

    return _element

def set_key(_res, keys,val):
    for i,key in enumerate(keys):
        try:
            if i == len(keys) - 1:
                _res[key]=val
                _res = _res[key]
            else:
                _res = _res[key]
        except KeyError:
            _res[key]=val if i==len(keys)-1 else {}
            if i<len(keys):
                _res =_res[key]


class Validate:
    def __init__(self, form: dict, break_on_error: bool = False) -> None:
        self._form = copy.deepcopy(form)
        self._break_on_error = break_on_error
        self._error = False
        self._msg = {}

    def has_default(self, rules):
        for item in rules:
            if isinstance(item, vr.Default):
                return item
        return False

    def has_required(self, rules):
        for item in rules:
            if type(item) == type(vr.Required):
                return True
        return False

    def rules(self, rules_data: dict, msg: str = 'invalid_data') -> dictAttribute:
        res = {}
        for field in rules_data:
            keys=field.split('.')
            val = keys_exists(self._form, keys)
            default = self.has_default(rules_data[field])
            if val is None and default is not False:
                val = default(val)

            check = self.__check(val, rules_data[field])
            if check == True:
                set_key(res,keys,val)
            else:
                self._error = True
                self._msg[field] = check
                if self._break_on_error:
                    break

        abort_if(self._error, {'msg': msg, 'errors': self._msg}, 422)

        return dictAttribute(res)

    def __check(self, input, rules):
        error = False
        msg = []
        for rule in rules:
            if rule != '$or' and not isinstance(rule, vr.Default):
                test = rule(input)
                if test != True:
                    error = True
                    msg.append(test)
            elif rule == '$or':
                test = False
                for item in rules['$or']:
                    test = item(input)
                    if test == True:
                        break
                if test != True:
                    error = True
                    msg.append(test)

        return True if error == False else msg



import html
import re
import typing, json

from multipart.multipart import parse_options_header
from http import cookies as http_cookies
from urllib.parse import parse_qsl
from renus.core.config import Config
from renus.core.formparsers import FormParser, MultiPartParser
from renus.core.injection import Injection


class Request:
    def __init__(self, scope, receive) -> None:
        self._scope = scope
        self._receive = receive
        self._headers = headers_parser(self._scope.get('headers', []))
        self._stream_consumed = False
        self.inputs = {}
        self.route = {}

    @property
    def headers(self) -> dict:
        return self._headers

    @property
    def cookies(self):
        if not hasattr(self, "_cookies"):
            headers = self.headers
            self._cookies = {}
            if 'cookie' in headers:
                self._cookies = cookie_parser(headers['cookie'])

        return self._cookies

    @property
    def client(self):
        return self._scope.get("client", [])

    @property
    def ip(self):
        return self.headers.get(Config('app').get('ipHeader', False), self.client[0])

    @property
    def sub_domain(self):
        if not hasattr(self, "_sub_domain"):
            host = self.headers.get('host', None)
            if host is None:
                return None

            ip = re.findall(r'[0-9]+(?:\.[0-9]+){3}', host)
            if len(ip) != 0:
                return None
            r = host.split('.')

            if len(r) <= 1:
                return None
            if len(r) == 2:
                if r[1].split(':')[0] == 'localhost':
                    return r[0]
                else:
                    return None
            self._sub_domain = r[0]

        return self._sub_domain

    @property
    def user_agent(self):
        return self.headers.get("user-agent", None)

    @property
    def base_path(self):
        if not hasattr(self, "_base_path"):
            headers = self.headers
            scheme = self._scope.get("scheme", "http")
            server = self._scope.get("server", None)
            host_header = None
            if 'host' in headers:
                host_header = headers['host']

            if host_header is not None:
                url = f"{scheme}://{host_header}"
            elif server is None:
                url = ''
            else:
                host, port = server
                default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[scheme]
                if port == default_port:
                    url = f"{scheme}://{host}"
                else:
                    url = f"{scheme}://{host}:{port}"

            self._base_path = url

        return self._base_path

    @property
    def full_path(self):
        if not hasattr(self, "_full_path"):
            path = self._scope.get("root_path", "") + self._scope.get("path", "")
            query_string = self._scope.get("query_string", b"")
            url = self.base_path
            url += path
            if query_string:
                url += "?" + query_string.decode()
            self._full_path = url

        return self._full_path

    @property
    def query_params(self):
        if not hasattr(self, "_query_params"):
            self._query_params = query_parser(self._scope.get('query_string', ''))
        return self._query_params

    @property
    def method(self) -> str:
        return self._scope.get("method", '')

    async def stream(self) -> typing.AsyncGenerator[bytes, None]:
        if hasattr(self, "_body"):
            yield self._body
            yield b""
            return

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        self._stream_consumed = True
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                body = message.get("body", b"")
                if body:
                    yield body
                if not message.get("more_body", False):
                    break
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()
        yield b""

    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            chunks = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)
        return self._body

    async def form(self):
        if not hasattr(self, "_form"):
            content_type_header = self.headers.get("content-type")
            content_type, options = parse_options_header(content_type_header)
            if content_type == b"multipart/form-data":
                multipart_parser = MultiPartParser(self.headers, self.stream())
                self._form = await multipart_parser.parse()
            elif content_type == b"app/x-www-form-urlencoded":
                form_parser = FormParser(self.headers, self.stream())
                self._form = await form_parser.parse()
            else:
                body = await self.body()
                try:
                    self._form = {} if body == b"" else json.loads(body)
                except Exception:
                    self._form = {}

                self._form = Injection().protect(self._form)

        return self._form


def headers_parser(headers: list) -> typing.Dict[str, str]:
    headers_dict: typing.Dict[str, str] = {}
    for header in headers:
        headers_dict[html.escape(str(header[0].decode("utf-8")).lower())] = html.escape(header[1].decode("utf-8"))

    return headers_dict


def cookie_parser(cookie_string: str) -> typing.Dict[str, str]:
    cookie_dict: typing.Dict[str, str] = {}
    cookie_string = html.escape(cookie_string)
    for chunk in cookie_string.split(";"):
        if "=" in chunk:
            key, val = chunk.split("=", 1)
        else:
            key, val = "", chunk
        key, val = key.strip(), val.strip()
        if key or val:
            cookie_dict[key] = http_cookies._unquote(val)
    return cookie_dict


def query_parser(query_string: str):
    params = parse_qsl(query_string.decode("utf-8"), keep_blank_values=True)
    params = {k: v for k, v in params}
    params = Injection().protect(params)
    return params


class ClientDisconnect(Exception):
    pass



from base64 import urlsafe_b64encode, urlsafe_b64decode
import sys

from renus.util.helper import get_random_string, remove_pad64, add_pad64

try:
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    from cryptography.fernet import Fernet
except ImportError:
    Fernet = None

def pad(s,n=16):
    block_size = n
    remainder = len(s) % block_size
    padding_needed = block_size - remainder
    return s + padding_needed * ' '


def __generate_key(password):
    salt = b'dfkTGG536fdfkTGG536fkdfijdkdfijd'
    kdf = PBKDF2HMAC(algorithm = hashes.SHA256(),length = 32,salt = salt,iterations = 100000)
    return urlsafe_b64encode(kdf.derive(password.encode()))


def encrypt(text:str,password:str)->str:
    assert Fernet is not None, "'cryptography' must be installed to use encrypt"
    key = __generate_key(password)
    f = Fernet(key)
    d= f.encrypt(text.encode())
    return d.decode()



def decrypt(enc_dict:str, password:str)->str:
    assert Fernet is not None, "'cryptography' must be installed to use decrypt"
    key = __generate_key(password)
    f = Fernet(key)
    d = f.decrypt(enc_dict.encode())
    return d.decode()


def shift(text: str, password: str):
    text += password
    rnd = get_random_string(6)
    password = rnd + password
    password = urlsafe_b64encode(password.encode()).decode()
    text = urlsafe_b64encode(text.encode()).decode()
    ls = []
    ln = len(password)
    for i in range(len(text)):
        key_c = password[i % ln]
        ls.append(chr((ord(text[i]) + ord(key_c)) % 256))

    text = rnd + ''.join(ls)
    text = urlsafe_b64encode(text.encode()).decode()
    return remove_pad64(text)


def unshift(text: str, password: str):
    try:
        lnpass = len(password)
        text = add_pad64(text)
        text = urlsafe_b64decode(text).decode()
        rnd = text[:6]
        text = text[6:]
        password = rnd + password
        password = urlsafe_b64encode(password.encode()).decode()
        ls = []
        ln = len(password)
        for i in range(len(text)):
            key_c = password[i % ln]
            ls.append(chr((256 + ord(text[i]) - ord(key_c)) % 256))

        text = ''.join(ls).rstrip()
        return urlsafe_b64decode(text).decode()[:-lnpass]
    except:
        raise RuntimeError('Invalid Token Format')


import os
import typing

from bson import ObjectId
from pymongo import MongoClient
from datetime import datetime
from renus.core.config import Config
from renus.core.request import Request
from renus.util.helper import dictAttribute

class ModelBase:
    _client = MongoClient(Config('database').get('host', '127.0.0.1'),
                          Config('database').get('port', 27017),
                          username=Config('database').get('username', None),
                          password=Config('database').get('password', None))

    _database_name = Config('database').get('name', 'renus')
    _collection_name=None
    metro = None
    _public_folder='public'
    hidden_fields = []
    fields = {}

    def __init__(self, request: Request) -> None:
        self._request = request
        self._steps = None
        self._where = None
        self._distinct = None
        self._limit = None
        self._skip = None
        self._sort = None
        self._select = None
        self.visible_fields = []

    def __iter__(self):
        return iter(self.get())

    def set_database(self, name: str):
        self._database_name = name
        return self

    def set_collection(self, name: str):
        self._collection_name = name
        return self

    @property
    def db(self):
        return self._client[self._database_name]

    def collection(self, name: str = None):
        if name is None:
            name = self._collection_name

        return self.db[name]

    def reset(self):
        self._steps = None
        self._where = None
        self._distinct = None
        self._limit = None
        self._skip = None
        self._sort = None
        self._select = None
        self.visible_fields = []
        return self

    def create_index(self, fields, unique=False):
        return self.collection().create_index(fields,
                                              unique=unique)

    def aggregate(self, pipeline: typing.Any, session: typing.Any = None):
        return self.__police(self.collection().aggregate(pipeline, session))

    def where(self, where: dict):
        if self._steps is not None:
            self._steps.append({
                "$match": where
            })
        else:
            if self._where is None:
                self._where = {}
            for k,v in where.items():
                if k.startswith('$') and k in self._where:
                    if type(v) is list:
                        self._where[k]+=v
                    else:
                        self._where[k]={**self._where[k],**v}
                else:
                    self._where[k] = v

        return self

    def count(self, skip=False, limit=False):
        k = {}
        if self._where is not None:
            k['filter'] = self._where
        else :
            k['filter'] = {}
        if skip and self._skip is not None:
            k['skip'] = self._skip
        if limit and self._limit is not None:
            k['limit'] = self._limit
        
        return self.collection().count_documents(**k)

    def distinct(self, key: str):
        self._distinct = key
        return self

    def limit(self, count: int):
        if self._steps is not None:
            self._steps.append({"$limit": count})
        else:
            self._limit = count
        return self

    def skip(self, count: int):
        if self._steps is not None:
            self._steps.append({"$skip": count})
        else:
            self._skip = count
        return self

    def group(self, item: dict):
        if self._steps is None:
            raise RuntimeError("group work with steps() architect")
        self._steps.append({"$group": item})
        return self

    def sort(self, key_or_list: typing.Union[str, typing.List[typing.Tuple]], asc: bool = True):
        if type(key_or_list) is dict:
            s = key_or_list
        elif type(key_or_list) is list:
            s = []
            for f, a in key_or_list:
                s.append((f, 1 if a else -1))
        else:
            s = [(key_or_list, 1 if asc else -1)]
        if self._steps is not None:
            if type(s) is list:
                s = {i[0]: i[1] for i in s}
            self._steps.append({"$sort": s})
        else:
            self._sort=s
        return self

    def select(self, *select: [str, typing.Dict]):
        s = {}
        if len(select)==1 and type(select[0]) is dict:
            s=select[0]
        else:
            for key in select:
                if type(key) is dict:
                    s = key
                else:
                    s[key] = 1

        if self._steps is not None:
            self._steps.append({'$project': s})
            return self

        self._select=s
        return self

    def with_relation(self, collection, local_field, forigen_field, to,single=False):
        if self._steps is not None:
            self._steps.append({
                "$lookup": {
                    "from": collection,
                    "localField": local_field,
                    "foreignField": forigen_field,
                    "as": to
                }
            })
            if single:
                self._steps.append({'$unwind': f'${to}'})
        return self

    def steps(self):
        self._steps = []
        return self

    def get(self, police: bool = True) -> typing.List[dictAttribute]:
        if self._steps is not None:
            return self.aggregate(self._steps)

        where, select = self._base_gate(police)

        find = self.collection().find(where, select)

        if self._sort is not None:
            find = find.sort(self._sort)
        if self._skip is not None:
            find = find.skip(self._skip)
        if self._limit is not None:
            find = find.limit(self._limit)
        if self._distinct is not None:
            return find.distinct(self._distinct)

        return self.__police(find) if police else list(find)

    def first(self, police: bool = True) -> typing.Union[None, dictAttribute]:
        self.limit(1)
        res = self.get(police)
        self._limit = None
        if len(res) == 1:
            return res[0]
        return None

    def create(self, document: dict) -> dict:
        if "updated_at" not in document:
            document["updated_at"] = datetime.utcnow()

        if "created_at" not in document:
            document["created_at"] = datetime.utcnow()

        id = self.collection().insert_one(document).inserted_id
        document['_id'] = id
        self.boot_event('create', {}, document)
        return document

    def create_many(self, documents: typing.List) -> typing.List[ObjectId]:
        for document in documents:
            if "updated_at" not in document:
                document["updated_at"] = datetime.utcnow()

            if "created_at" not in document:
                document["created_at"] = datetime.utcnow()

        ids = self.collection().insert_many(documents).inserted_ids
        self.boot_event('create_many', {}, ids)
        return ids

    def update(self, new: dict, upsert=False, get_old=False) -> bool:
        where = self.__ud_gate('update')
        new["updated_at"] = datetime.utcnow()
        d = {"$set": new}
        if "created_at" not in new:
            d["$setOnInsert"] = {'created_at': datetime.utcnow()}
        old = self.collection().find_one_and_update(where, d, upsert=upsert)
        self.boot_event('update', old, new)
        return old if get_old else True

    def update_opt(self, new: dict, upsert=False, get_old=False) -> bool:
        where = self.__ud_gate('update')
        if '$set' not in new:
            new['$set'] = {}
        if '$setOnInsert' not in new:
            new['$setOnInsert'] = {}
        new['$set']["updated_at"] = datetime.utcnow()
        new['$setOnInsert']['created_at'] = datetime.utcnow()
        old = self.collection().find_one_and_update(where, new, upsert=upsert)
        self.boot_event('update', old, {k[1:]:v for k,v in new.items()})
        return old if get_old else True

    def update_opt_many(self, new: dict, upsert=False, get_old=False) -> bool:
        where = self.__ud_gate('update')
        if '$set' not in new:
            new['$set'] = {}
        if '$setOnInsert' not in new:
            new['$setOnInsert'] = {}
        new['$set']["updated_at"] = datetime.utcnow()
        new['$setOnInsert']['created_at'] = datetime.utcnow()
        old = self.collection().update_many(where, new, upsert=upsert).raw_result
        self.boot_event('update', old, {k[1:]:v for k,v in new.items()})
        return old if get_old else True

    def update_many(self, new: dict) -> bool:
        where = self.__ud_gate('update')
        new["updated_at"] = datetime.utcnow()
        old = self.collection().update_many(where, {"$set": new}).raw_result
        self.boot_event('update_many', old, str(new))
        return True

    def delete(self, all: bool = False) -> bool:
        where = self.__ud_gate('delete')
        if all is False:
            old = self.collection().find_one_and_delete(where)
            if self.metro is not None:
                self._handle_metro('delete', old)
            self.boot_event('delete', old, {})
        else:
            if self.metro is not None:
                self._handle_metro('delete')
            old = self.collection().delete_many(where)

            self.boot_event('delete_many', {'deleted_count': old.deleted_count, 'where': str(where)}, {})

        return old

    def make_visible(self, fields: typing.List):
        self.visible_fields=fields
        return self

    @staticmethod
    def boot_event(typ: str, old, new):
        pass

    @staticmethod
    def convert_id(id):
        if type(id) is str:
            try:
                return ObjectId(id)
            except Exception:
                return id
        return id

    def _base_gate(self, police) -> list:
        where = self._where
        select = self._select

        if select is None and police == False:
            raise RuntimeError("when police is off, for security reason select fields. ex: .select('_id','name')")

        if where is not None:
            if '_id' in where:
                where['_id'] = self.convert_id(where['_id'])

        return [where, select]

    @staticmethod
    def cast(document:dict):
        return document
    
    def __cleaner(self, document: dict) -> dictAttribute:
        for field in self.hidden_fields:
            if field in document and field not in self.visible_fields:
                del document[field]
        
        return dictAttribute(self.cast(document))

    def __police(self, documents: typing.Any):
        if documents is None:
            return None
        res = []
        if type(documents) is dict:
            return self.__cleaner(documents)
        else:
            for document in documents:
                if type(document) is dict:
                    res.append(self.__cleaner(document))
        return res

    def __ud_gate(self, type: str):
        where = self._where
        if where is None:
            raise RuntimeError(f"for {type} use where. ex: .where({'name':'test'})")
        if '_id' in where:
            where['_id'] = self.convert_id(where['_id'])

        return where

    def _handle_metro(self, typ, obj=None):
        if typ == 'delete':
            if obj is None:
                where = self.__ud_gate('delete')
                all = self.collection().find(where)
            else:
                all = [obj]

            for item in all:
                for field, db in self.metro.items():
                    if type(db) is not list:
                        raise RuntimeError("metro format not true. ex: '_id':[{'collection':'test','field': 'test_id'}]")
                    for d in db:
                        strg=d.get('storage',False)
                        c=d.get('collection',False)
                        if c is not False:
                            if strg:
                                files = self.collection(c).find({
                                    d['field']: item[field]
                                })

                                for file in files:
                                    self._remove_file(file.get(strg, []))
                            self.collection(c).delete_many({
                                d['field']: item[field]
                            })
                        elif strg is not False:
                            l = self.__get_links(strg, item)
                            if l:
                                self._remove_file(l)


    def __get_links(self,field:str,doc:dict):
        item = doc.copy()
        p = field.split('.')
        r = None
        for i in p:
            if i not in item:
                return False
            r = item[i]
            item = item[i]
        return r

    def _remove_file(self, links):
        if type(links) is dict:
            links = list(links.values())
        if type(links) is not list:
            links = [links]

        for link in links:
            if type(link) is str:
                link = link.replace('storage/img/', '', 1)
                link = link.replace('storage/', '', 1)
            elif type(link) is dict:
                link = link.get('url').replace('storage/img/', '', 1)
                link = link.replace('storage/', '', 1)
            else:
                raise RuntimeError(f"type {link} must be string or dict but its {type(link)}")

            try:
                os.remove(f'storage/{self._public_folder}/' + link)
            except:
                pass

    @property
    def database_name(self):
        return self._database_name
